@@ -1,4 +1,3 @@
-/* vim:set ft=c ts=2 sw=2 sts=2 et cindent: */
 /*
  * Portions created by Alan Antonuk are Copyright (c) 2012-2014 Alan Antonuk.
  * All Rights Reserved.
@@ -29,65 +28,81 @@
 #include "config.h"
 #endif
 
-#include "amqp_ssl_socket.h"
-#include "amqp_socket.h"
-#include "amqp_hostcheck.h"
+#ifdef _MSC_VER
+#define _CRT_SECURE_NO_WARNINGS
+#endif
+
+#include "amqp_openssl_bio.h"
+#include "amqp_openssl_hostname_validation.h"
 #include "amqp_private.h"
+#include "amqp_socket.h"
+#include "amqp_ssl_socket.h"
 #include "amqp_time.h"
 #include "threads.h"
 
 #include <ctype.h>
+#include <limits.h>
+#include <openssl/bio.h>
 #include <openssl/conf.h>
+#include <openssl/engine.h>
 #include <openssl/err.h>
 #include <openssl/ssl.h>
 #include <openssl/x509v3.h>
 #include <stdlib.h>
 #include <string.h>
 
+static int initialize_ssl_and_increment_connections(void);
+static int decrement_ssl_connections(void);
 
-static int initialize_openssl(void);
-static int destroy_openssl(void);
+static unsigned long ssl_threadid_callback(void);
+static void ssl_locking_callback(int mode, int n, const char *file, int line);
+static pthread_mutex_t *amqp_openssl_lockarray = NULL;
 
-static int open_ssl_connections = 0;
+static pthread_mutex_t openssl_init_mutex = PTHREAD_MUTEX_INITIALIZER;
 static amqp_boolean_t do_initialize_openssl = 1;
 static amqp_boolean_t openssl_initialized = 0;
+static amqp_boolean_t openssl_bio_initialized = 0;
+static int openssl_connections = 0;
 
-#ifdef ENABLE_THREAD_SAFETY
-static unsigned long amqp_ssl_threadid_callback(void);
-static void amqp_ssl_locking_callback(int mode, int n, const char *file, int line);
-
-#ifdef _WIN32
-static long win32_create_mutex = 0;
-static pthread_mutex_t openssl_init_mutex = NULL;
-#else
-static pthread_mutex_t openssl_init_mutex = PTHREAD_MUTEX_INITIALIZER;
-#endif
-static pthread_mutex_t *amqp_openssl_lockarray = NULL;
-#endif /* ENABLE_THREAD_SAFETY */
+#define CHECK_SUCCESS(condition)                                            \
+  do {                                                                      \
+    int check_success_ret = (condition);                                    \
+    if (check_success_ret) {                                                \
+      amqp_abort("Check %s failed <%d>: %s", #condition, check_success_ret, \
+                 strerror(check_success_ret));                              \
+    }                                                                       \
+  } while (0)
 
 struct amqp_ssl_socket_t {
   const struct amqp_socket_class_t *klass;
   SSL_CTX *ctx;
   int sockfd;
   SSL *ssl;
-  amqp_boolean_t verify;
+  amqp_boolean_t verify_peer;
+  amqp_boolean_t verify_hostname;
   int internal_error;
 };
 
 static ssize_t amqp_ssl_socket_send(void *base, const void *buf, size_t len,
                                     AMQP_UNUSED int flags) {
   struct amqp_ssl_socket_t *self = (struct amqp_ssl_socket_t *)base;
-  ssize_t res;
+  int res;
   if (-1 == self->sockfd) {
     return AMQP_STATUS_SOCKET_CLOSED;
   }
 
+  /* SSL_write takes an int for length of buffer, protect against len being
+   * larger than larger than what SSL_write can take */
+  if (len > INT_MAX) {
+    return AMQP_STATUS_INVALID_PARAMETER;
+  }
+
   ERR_clear_error();
   self->internal_error = 0;
 
   /* This will only return on error, or once the whole buffer has been
    * written to the SSL stream. See SSL_MODE_ENABLE_PARTIAL_WRITE */
-  res = SSL_write(self->ssl, buf, len);
+  res = SSL_write(self->ssl, buf, (int)len);
   if (0 >= res) {
     self->internal_error = SSL_get_error(self->ssl, res);
     /* TODO: Close connection if it isn't already? */
@@ -110,156 +125,56 @@ static ssize_t amqp_ssl_socket_send(void *base, const void *buf, size_t len,
     self->internal_error = 0;
   }
 
-  return res;
+  return (ssize_t)res;
 }
 
-static ssize_t
-amqp_ssl_socket_recv(void *base,
-                     void *buf,
-                     size_t len,
-                     AMQP_UNUSED int flags)
-{
+static ssize_t amqp_ssl_socket_recv(void *base, void *buf, size_t len,
+                                    AMQP_UNUSED int flags) {
   struct amqp_ssl_socket_t *self = (struct amqp_ssl_socket_t *)base;
-  ssize_t received;
+  int received;
   if (-1 == self->sockfd) {
     return AMQP_STATUS_SOCKET_CLOSED;
   }
+
+  /* SSL_read takes an int for length of buffer, protect against len being
+   * larger than larger than what SSL_read can take */
+  if (len > INT_MAX) {
+    return AMQP_STATUS_INVALID_PARAMETER;
+  }
+
   ERR_clear_error();
   self->internal_error = 0;
 
-  received = SSL_read(self->ssl, buf, len);
+  received = SSL_read(self->ssl, buf, (int)len);
   if (0 >= received) {
     self->internal_error = SSL_get_error(self->ssl, received);
     switch (self->internal_error) {
-    case SSL_ERROR_WANT_READ:
+      case SSL_ERROR_WANT_READ:
         received = AMQP_PRIVATE_STATUS_SOCKET_NEEDREAD;
         break;
-    case SSL_ERROR_WANT_WRITE:
+      case SSL_ERROR_WANT_WRITE:
         received = AMQP_PRIVATE_STATUS_SOCKET_NEEDWRITE;
         break;
-    case SSL_ERROR_ZERO_RETURN:
-      received = AMQP_STATUS_CONNECTION_CLOSED;
-      break;
-    default:
-      received = AMQP_STATUS_SSL_ERROR;
-      break;
+      case SSL_ERROR_ZERO_RETURN:
+        received = AMQP_STATUS_CONNECTION_CLOSED;
+        break;
+      default:
+        received = AMQP_STATUS_SSL_ERROR;
+        break;
     }
   }
 
-  return received;
+  return (ssize_t)received;
 }
 
-static int match(ASN1_STRING *entry_string, const char *string)
-{
-  unsigned char *utf8_value = NULL, *cp, ch;
-  int utf8_length, status = 1;
-  utf8_length = ASN1_STRING_to_UTF8(&utf8_value, entry_string);
-  if (0 > utf8_length) {
-    goto error;
-  }
-  while (utf8_length > 0 && utf8_value[utf8_length - 1] == 0) {
-    --utf8_length;
-  }
-  if (utf8_length >= 256) {
-    goto error;
-  }
-  if ((size_t)utf8_length != strlen((char *)utf8_value)) {
-    goto error;
-  }
-  for (cp = utf8_value; (ch = *cp) != '\0'; ++cp) {
-    if (isascii(ch) && !isprint(ch)) {
-      goto error;
-    }
-  }
-  if (!amqp_hostcheck((char *)utf8_value, string)) {
-    goto error;
-  }
-exit:
-  OPENSSL_free(utf8_value);
-  return status;
-error:
-  status = 0;
-  goto exit;
-}
-
-/* Does this hostname match an entry in the subjectAltName extension?
- * returns: 0 if no, 1 if yes, -1 if no subjectAltName entries were found.
- */
-static int hostname_matches_subject_alt_name(const char *hostname, X509 *cert)
-{
-  int found_any_entries = 0;
-  int found_match;
-  GENERAL_NAME *namePart = NULL;
-  STACK_OF(GENERAL_NAME) *san =
-    (STACK_OF(GENERAL_NAME)*) X509_get_ext_d2i(cert, NID_subject_alt_name, NULL, NULL);
-
-  while (sk_GENERAL_NAME_num(san) > 0)
-  {
-    namePart = sk_GENERAL_NAME_pop(san);
-
-    if (namePart->type == GEN_DNS) {
-      found_any_entries = 1;
-      found_match = match(namePart->d.uniformResourceIdentifier, hostname);
-      if (found_match)
-        return 1;
-    }
-  }
-
-  return (found_any_entries ? 0 : -1);
-}
-
-static int hostname_matches_subject_common_name(const char *hostname, X509 *cert)
-{
-  X509_NAME *name;
-  X509_NAME_ENTRY *name_entry;
-  int position;
-  ASN1_STRING *entry_string;
-
-  name = X509_get_subject_name(cert);
-  position = -1;
-  for (;;) {
-    position = X509_NAME_get_index_by_NID(name, NID_commonName, position);
-    if (position == -1)
-      break;
-    name_entry = X509_NAME_get_entry(name, position);
-    entry_string = X509_NAME_ENTRY_get_data(name_entry);
-    if (match(entry_string, hostname))
-      return 1;
-  }
-  return 0;
-}
-
-static int
-amqp_ssl_socket_verify_hostname(void *base, const char *host)
-{
-  struct amqp_ssl_socket_t *self = (struct amqp_ssl_socket_t *)base;
-  int status = 0;
-  X509 *cert;
-  int res;
-  cert = SSL_get_peer_certificate(self->ssl);
-  if (!cert) {
-    goto error;
-  }
-  res = hostname_matches_subject_alt_name(host, cert);
-  if (res != 1) {
-    res = hostname_matches_subject_common_name(host, cert);
-    if (!res)
-      goto error;
-  }
-exit:
-  return status;
-error:
-  status = -1;
-  goto exit;
-}
-
-static int
-amqp_ssl_socket_open(void *base, const char *host, int port, struct timeval *timeout)
-{
+static int amqp_ssl_socket_open(void *base, const char *host, int port,
+                                const struct timeval *timeout) {
   struct amqp_ssl_socket_t *self = (struct amqp_ssl_socket_t *)base;
   long result;
   int status;
   amqp_time_t deadline;
+  X509 *cert;
+  BIO *bio;
   if (-1 != self->sockfd) {
     return AMQP_STATUS_SOCKET_INUSE;
   }
@@ -285,7 +200,16 @@ amqp_ssl_socket_open(void *base, const char *host, int port, struct timeval *tim
     goto error_out1;
   }
 
-  status = SSL_set_fd(self->ssl, self->sockfd);
+  bio = BIO_new(amqp_openssl_bio());
+  if (!bio) {
+    status = AMQP_STATUS_NO_MEMORY;
+    goto error_out2;
+  }
+
+  BIO_set_fd(bio, self->sockfd, BIO_NOCLOSE);
+  SSL_set_bio(self->ssl, bio, bio);
+
+  status = SSL_set_tlsext_host_name(self->ssl, host);
   if (!status) {
     self->internal_error = SSL_get_error(self->ssl, status);
     status = AMQP_STATUS_SSL_ERROR;
@@ -312,27 +236,45 @@ start_connect:
     goto error_out2;
   }
 
-  result = SSL_get_verify_result(self->ssl);
-  if (X509_V_OK != result) {
-    self->internal_error = result;
-    status = AMQP_STATUS_SSL_PEER_VERIFY_FAILED;
-    goto error_out3;
+  cert = SSL_get_peer_certificate(self->ssl);
+
+  if (self->verify_peer) {
+    if (!cert) {
+      self->internal_error = 0;
+      status = AMQP_STATUS_SSL_PEER_VERIFY_FAILED;
+      goto error_out3;
+    }
+
+    result = SSL_get_verify_result(self->ssl);
+    if (X509_V_OK != result) {
+      self->internal_error = result;
+      status = AMQP_STATUS_SSL_PEER_VERIFY_FAILED;
+      goto error_out4;
+    }
   }
-  if (self->verify) {
-    int verify_status = amqp_ssl_socket_verify_hostname(self, host);
-    if (verify_status) {
+  if (self->verify_hostname) {
+    if (!cert) {
       self->internal_error = 0;
       status = AMQP_STATUS_SSL_HOSTNAME_VERIFY_FAILED;
       goto error_out3;
     }
+
+    if (AMQP_HVR_MATCH_FOUND != amqp_ssl_validate_hostname(host, cert)) {
+      self->internal_error = 0;
+      status = AMQP_STATUS_SSL_HOSTNAME_VERIFY_FAILED;
+      goto error_out4;
+    }
   }
 
+  X509_free(cert);
   self->internal_error = 0;
   status = AMQP_STATUS_OK;
 
 exit:
   return status;
 
+error_out4:
+  X509_free(cert);
 error_out3:
   SSL_shutdown(self->ssl);
 error_out2:
@@ -344,35 +286,18 @@ error_out1:
   goto exit;
 }
 
-static int
-amqp_ssl_socket_close(void *base)
-{
-  int res;
+static int amqp_ssl_socket_close(void *base, amqp_socket_close_enum force) {
   struct amqp_ssl_socket_t *self = (struct amqp_ssl_socket_t *)base;
 
   if (-1 == self->sockfd) {
     return AMQP_STATUS_SOCKET_CLOSED;
   }
 
-start_shutdown:
-  res = SSL_shutdown(self->ssl);
-  if (0 == res) {
-    goto start_shutdown;
-  } else if (-1 == res) {
-    self->internal_error = SSL_get_error(self->ssl, res);
-    switch (self->internal_error) {
-      case SSL_ERROR_WANT_READ:
-        res = amqp_poll(self->sockfd, AMQP_SF_POLLIN, amqp_time_infinite());
-        break;
-      case SSL_ERROR_WANT_WRITE:
-        res = amqp_poll(self->sockfd, AMQP_SF_POLLOUT, amqp_time_infinite());
-        break;
-    }
-    if (AMQP_STATUS_OK == res) {
-      goto start_shutdown;
-    }
-    /* Swallow errors in poll, just consider the connection dead */
+  if (AMQP_SC_NONE == force) {
+    /* don't try too hard to shutdown the connection */
+    SSL_shutdown(self->ssl);
   }
+
   SSL_free(self->ssl);
   self->ssl = NULL;
 
@@ -384,39 +309,33 @@ start_shutdown:
   return AMQP_STATUS_OK;
 }
 
-static int
-amqp_ssl_socket_get_sockfd(void *base)
-{
+static int amqp_ssl_socket_get_sockfd(void *base) {
   struct amqp_ssl_socket_t *self = (struct amqp_ssl_socket_t *)base;
   return self->sockfd;
 }
 
-static void
-amqp_ssl_socket_delete(void *base)
-{
+static void amqp_ssl_socket_delete(void *base) {
   struct amqp_ssl_socket_t *self = (struct amqp_ssl_socket_t *)base;
 
   if (self) {
-    amqp_ssl_socket_close(self);
+    amqp_ssl_socket_close(self, AMQP_SC_NONE);
 
     SSL_CTX_free(self->ctx);
     free(self);
   }
-  destroy_openssl();
+  decrement_ssl_connections();
 }
 
 static const struct amqp_socket_class_t amqp_ssl_socket_class = {
-  amqp_ssl_socket_send, /* send */
-  amqp_ssl_socket_recv, /* recv */
-  amqp_ssl_socket_open, /* open */
-  amqp_ssl_socket_close, /* close */
-  amqp_ssl_socket_get_sockfd, /* get_sockfd */
-  amqp_ssl_socket_delete /* delete */
+    amqp_ssl_socket_send,       /* send */
+    amqp_ssl_socket_recv,       /* recv */
+    amqp_ssl_socket_open,       /* open */
+    amqp_ssl_socket_close,      /* close */
+    amqp_ssl_socket_get_sockfd, /* get_sockfd */
+    amqp_ssl_socket_delete      /* delete */
 };
 
-amqp_socket_t *
-amqp_ssl_socket_new(amqp_connection_state_t state)
-{
+amqp_socket_t *amqp_ssl_socket_new(amqp_connection_state_t state) {
   struct amqp_ssl_socket_t *self = calloc(1, sizeof(*self));
   int status;
   if (!self) {
@@ -425,9 +344,10 @@ amqp_ssl_socket_new(amqp_connection_state_t state)
 
   self->sockfd = -1;
   self->klass = &amqp_ssl_socket_class;
-  self->verify = 1;
+  self->verify_peer = 1;
+  self->verify_hostname = 1;
 
-  status = initialize_openssl();
+  status = initialize_ssl_and_increment_connections();
   if (status) {
     goto error;
   }
@@ -436,6 +356,13 @@ amqp_ssl_socket_new(amqp_connection_state_t state)
   if (!self->ctx) {
     goto error;
   }
+  /* Disable SSLv2 and SSLv3 */
+  SSL_CTX_set_options(self->ctx, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
+
+  SSL_CTX_set_mode(self->ctx, SSL_MODE_ENABLE_PARTIAL_WRITE);
+  /* OpenSSL v1.1.1 turns this on by default, which makes the non-blocking
+   * logic not behave as expected, so turn this back off */
+  SSL_CTX_clear_mode(self->ctx, SSL_MODE_AUTO_RETRY);
 
   amqp_set_socket(state, (amqp_socket_t *)self);
 
@@ -445,10 +372,14 @@ error:
   return NULL;
 }
 
-int
-amqp_ssl_socket_set_cacert(amqp_socket_t *base,
-                           const char *cacert)
-{
+void *amqp_ssl_socket_get_context(amqp_socket_t *base) {
+  if (base->klass != &amqp_ssl_socket_class) {
+    amqp_abort("<%p> is not of type amqp_ssl_socket_t", base);
+  }
+  return ((struct amqp_ssl_socket_t *)base)->ctx;
+}
+
+int amqp_ssl_socket_set_cacert(amqp_socket_t *base, const char *cacert) {
   int status;
   struct amqp_ssl_socket_t *self;
   if (base->klass != &amqp_ssl_socket_class) {
@@ -462,11 +393,8 @@ amqp_ssl_socket_set_cacert(amqp_socket_t *base,
   return AMQP_STATUS_OK;
 }
 
-int
-amqp_ssl_socket_set_key(amqp_socket_t *base,
-                        const char *cert,
-                        const char *key)
-{
+int amqp_ssl_socket_set_key(amqp_socket_t *base, const char *cert,
+                            const char *key) {
   int status;
   struct amqp_ssl_socket_t *self;
   if (base->klass != &amqp_ssl_socket_class) {
@@ -477,30 +405,20 @@ amqp_ssl_socket_set_key(amqp_socket_t *base,
   if (1 != status) {
     return AMQP_STATUS_SSL_ERROR;
   }
-  status = SSL_CTX_use_PrivateKey_file(self->ctx, key,
-                                       SSL_FILETYPE_PEM);
+  status = SSL_CTX_use_PrivateKey_file(self->ctx, key, SSL_FILETYPE_PEM);
   if (1 != status) {
     return AMQP_STATUS_SSL_ERROR;
   }
   return AMQP_STATUS_OK;
 }
 
-static int
-password_cb(AMQP_UNUSED char *buffer,
-            AMQP_UNUSED int length,
-            AMQP_UNUSED int rwflag,
-            AMQP_UNUSED void *user_data)
-{
+static int password_cb(AMQP_UNUSED char *buffer, AMQP_UNUSED int length,
+                       AMQP_UNUSED int rwflag, AMQP_UNUSED void *user_data) {
   amqp_abort("rabbitmq-c does not support password protected keys");
-  return 0;
 }
 
-int
-amqp_ssl_socket_set_key_buffer(amqp_socket_t *base,
-                               const char *cert,
-                               const void *key,
-                               size_t n)
-{
+int amqp_ssl_socket_set_key_buffer(amqp_socket_t *base, const char *cert,
+                                   const void *key, size_t n) {
   int status = AMQP_STATUS_OK;
   BIO *buf = NULL;
   RSA *rsa = NULL;
@@ -508,12 +426,15 @@ amqp_ssl_socket_set_key_buffer(amqp_socket_t *base,
   if (base->klass != &amqp_ssl_socket_class) {
     amqp_abort("<%p> is not of type amqp_ssl_socket_t", base);
   }
+  if (n > INT_MAX) {
+    return AMQP_STATUS_INVALID_PARAMETER;
+  }
   self = (struct amqp_ssl_socket_t *)base;
   status = SSL_CTX_use_certificate_chain_file(self->ctx, cert);
   if (1 != status) {
     return AMQP_STATUS_SSL_ERROR;
   }
-  buf = BIO_new_mem_buf((void *)key, n);
+  buf = BIO_new_mem_buf((void *)key, (int)n);
   if (!buf) {
     goto error;
   }
@@ -534,10 +455,7 @@ error:
   goto exit;
 }
 
-int
-amqp_ssl_socket_set_cert(amqp_socket_t *base,
-                         const char *cert)
-{
+int amqp_ssl_socket_set_cert(amqp_socket_t *base, const char *cert) {
   int status;
   struct amqp_ssl_socket_t *self;
   if (base->klass != &amqp_ssl_socket_class) {
@@ -551,140 +469,251 @@ amqp_ssl_socket_set_cert(amqp_socket_t *base,
   return AMQP_STATUS_OK;
 }
 
-void
-amqp_ssl_socket_set_verify(amqp_socket_t *base,
-                           amqp_boolean_t verify)
-{
+void amqp_ssl_socket_set_verify(amqp_socket_t *base, amqp_boolean_t verify) {
+  amqp_ssl_socket_set_verify_peer(base, verify);
+  amqp_ssl_socket_set_verify_hostname(base, verify);
+}
+
+void amqp_ssl_socket_set_verify_peer(amqp_socket_t *base,
+                                     amqp_boolean_t verify) {
   struct amqp_ssl_socket_t *self;
   if (base->klass != &amqp_ssl_socket_class) {
     amqp_abort("<%p> is not of type amqp_ssl_socket_t", base);
   }
   self = (struct amqp_ssl_socket_t *)base;
-  self->verify = verify;
+  self->verify_peer = verify;
 }
 
-void
-amqp_set_initialize_ssl_library(amqp_boolean_t do_initialize)
-{
-  if (!openssl_initialized) {
+void amqp_ssl_socket_set_verify_hostname(amqp_socket_t *base,
+                                         amqp_boolean_t verify) {
+  struct amqp_ssl_socket_t *self;
+  if (base->klass != &amqp_ssl_socket_class) {
+    amqp_abort("<%p> is not of type amqp_ssl_socket_t", base);
+  }
+  self = (struct amqp_ssl_socket_t *)base;
+  self->verify_hostname = verify;
+}
+
+int amqp_ssl_socket_set_ssl_versions(amqp_socket_t *base,
+                                     amqp_tls_version_t min,
+                                     amqp_tls_version_t max) {
+  struct amqp_ssl_socket_t *self;
+  if (base->klass != &amqp_ssl_socket_class) {
+    amqp_abort("<%p> is not of type amqp_ssl_socket_t", base);
+  }
+  self = (struct amqp_ssl_socket_t *)base;
+
+  {
+    long clear_options;
+    long set_options = 0;
+#if defined(SSL_OP_NO_TLSv1_2)
+    amqp_tls_version_t max_supported = AMQP_TLSv1_2;
+    clear_options = SSL_OP_NO_TLSv1 | SSL_OP_NO_TLSv1_1 | SSL_OP_NO_TLSv1_2;
+#elif defined(SSL_OP_NO_TLSv1_1)
+    amqp_tls_version_t max_supported = AMQP_TLSv1_1;
+    clear_options = SSL_OP_NO_TLSv1 | SSL_OP_NO_TLSv1_1;
+#elif defined(SSL_OP_NO_TLSv1)
+    amqp_tls_version_t max_supported = AMQP_TLSv1;
+    clear_options = SSL_OP_NO_TLSv1;
+#else
+#error "Need a version of OpenSSL that can support TLSv1 or greater."
+#endif
+
+    if (AMQP_TLSvLATEST == max) {
+      max = max_supported;
+    }
+    if (AMQP_TLSvLATEST == min) {
+      min = max_supported;
+    }
+
+    if (min > max) {
+      return AMQP_STATUS_INVALID_PARAMETER;
+    }
+
+    if (max > max_supported || min > max_supported) {
+      return AMQP_STATUS_UNSUPPORTED;
+    }
+
+    if (min > AMQP_TLSv1) {
+      set_options |= SSL_OP_NO_TLSv1;
+    }
+#ifdef SSL_OP_NO_TLSv1_1
+    if (min > AMQP_TLSv1_1 || max < AMQP_TLSv1_1) {
+      set_options |= SSL_OP_NO_TLSv1_1;
+    }
+#endif
+#ifdef SSL_OP_NO_TLSv1_2
+    if (max < AMQP_TLSv1_2) {
+      set_options |= SSL_OP_NO_TLSv1_2;
+    }
+#endif
+    SSL_CTX_clear_options(self->ctx, clear_options);
+    SSL_CTX_set_options(self->ctx, set_options);
+  }
+
+  return AMQP_STATUS_OK;
+}
+
+void amqp_set_initialize_ssl_library(amqp_boolean_t do_initialize) {
+  CHECK_SUCCESS(pthread_mutex_lock(&openssl_init_mutex));
+
+  if (openssl_connections == 0 && !openssl_initialized) {
     do_initialize_openssl = do_initialize;
   }
+  CHECK_SUCCESS(pthread_mutex_unlock(&openssl_init_mutex));
 }
 
-#ifdef ENABLE_THREAD_SAFETY
-unsigned long
-amqp_ssl_threadid_callback(void)
-{
+static unsigned long ssl_threadid_callback(void) {
   return (unsigned long)pthread_self();
 }
 
-void
-amqp_ssl_locking_callback(int mode, int n,
-                          AMQP_UNUSED const char *file,
-                          AMQP_UNUSED int line)
-{
+static void ssl_locking_callback(int mode, int n, AMQP_UNUSED const char *file,
+                                 AMQP_UNUSED int line) {
   if (mode & CRYPTO_LOCK) {
-    if (pthread_mutex_lock(&amqp_openssl_lockarray[n])) {
-      amqp_abort("Runtime error: Failure in trying to lock OpenSSL mutex");
-    }
+    CHECK_SUCCESS(pthread_mutex_lock(&amqp_openssl_lockarray[n]));
   } else {
-    if (pthread_mutex_unlock(&amqp_openssl_lockarray[n])) {
-      amqp_abort("Runtime error: Failure in trying to unlock OpenSSL mutex");
-    }
+    CHECK_SUCCESS(pthread_mutex_unlock(&amqp_openssl_lockarray[n]));
   }
 }
-#endif /* ENABLE_THREAD_SAFETY */
 
-static int
-initialize_openssl(void)
-{
-#ifdef ENABLE_THREAD_SAFETY
-#ifdef _WIN32
-  /* No such thing as PTHREAD_INITIALIZE_MUTEX macro on Win32, so we use this */
-  if (NULL == openssl_init_mutex) {
-    while (InterlockedExchange(&win32_create_mutex, 1) == 1)
-      /* Loop, someone else is holding this lock */ ;
+static int setup_openssl(void) {
+  int status;
 
-    if (NULL == openssl_init_mutex) {
-      if (pthread_mutex_init(&openssl_init_mutex, NULL)) {
-        return -1;
-      }
-    }
-    InterlockedExchange(&win32_create_mutex, 0);
+  int i;
+  amqp_openssl_lockarray = calloc(CRYPTO_num_locks(), sizeof(pthread_mutex_t));
+  if (!amqp_openssl_lockarray) {
+    status = AMQP_STATUS_NO_MEMORY;
+    goto out;
   }
-#endif /* _WIN32 */
-
-  if (pthread_mutex_lock(&openssl_init_mutex)) {
-    return -1;
-  }
-#endif /* ENABLE_THREAD_SAFETY */
-  if (do_initialize_openssl) {
-#ifdef ENABLE_THREAD_SAFETY
-    if (NULL == amqp_openssl_lockarray) {
-      int i = 0;
-      amqp_openssl_lockarray = calloc(CRYPTO_num_locks(), sizeof(pthread_mutex_t));
-      if (!amqp_openssl_lockarray) {
-        pthread_mutex_unlock(&openssl_init_mutex);
-        return -1;
+  for (i = 0; i < CRYPTO_num_locks(); i++) {
+    if (pthread_mutex_init(&amqp_openssl_lockarray[i], NULL)) {
+      int j;
+      for (j = 0; j < i; j++) {
+        pthread_mutex_destroy(&amqp_openssl_lockarray[j]);
       }
-      for (i = 0; i < CRYPTO_num_locks(); ++i) {
-        if (pthread_mutex_init(&amqp_openssl_lockarray[i], NULL)) {
-          free(amqp_openssl_lockarray);
-          amqp_openssl_lockarray = NULL;
-          pthread_mutex_unlock(&openssl_init_mutex);
-          return -1;
-        }
-      }
-    }
-
-    if (0 == open_ssl_connections) {
-      CRYPTO_set_id_callback(amqp_ssl_threadid_callback);
-      CRYPTO_set_locking_callback(amqp_ssl_locking_callback);
-    }
-#endif /* ENABLE_THREAD_SAFETY */
-
-    if (!openssl_initialized) {
-      OPENSSL_config(NULL);
-
-      SSL_library_init();
-      SSL_load_error_strings();
-
-      openssl_initialized = 1;
+      free(amqp_openssl_lockarray);
+      status = AMQP_STATUS_SSL_ERROR;
+      goto out;
     }
   }
+  CRYPTO_set_id_callback(ssl_threadid_callback);
+  CRYPTO_set_locking_callback(ssl_locking_callback);
 
-  ++open_ssl_connections;
+#ifdef AMQP_OPENSSL_V110
+  if (OPENSSL_init_ssl(0, NULL) <= 0) {
+    status = AMQP_STATUS_SSL_ERROR;
+    goto out;
+  }
+#else
+  OPENSSL_config(NULL);
+#endif
+  SSL_library_init();
+  SSL_load_error_strings();
 
-#ifdef ENABLE_THREAD_SAFETY
-  pthread_mutex_unlock(&openssl_init_mutex);
-#endif /* ENABLE_THREAD_SAFETY */
-  return 0;
+  status = AMQP_STATUS_OK;
+out:
+  return status;
 }
 
-static int
-destroy_openssl(void)
-{
-#ifdef ENABLE_THREAD_SAFETY
-  if (pthread_mutex_lock(&openssl_init_mutex)) {
-    return -1;
-  }
-#endif /* ENABLE_THREAD_SAFETY */
+int amqp_initialize_ssl_library(void) {
+  int status;
+  CHECK_SUCCESS(pthread_mutex_lock(&openssl_init_mutex));
 
-  if (open_ssl_connections > 0) {
-    --open_ssl_connections;
+  if (!openssl_initialized) {
+    status = setup_openssl();
+    if (status) {
+      goto out;
+    }
+    openssl_initialized = 1;
   }
 
-#ifdef ENABLE_THREAD_SAFETY
-  if (0 == open_ssl_connections && do_initialize_openssl) {
-    /* Unsetting these allows the rabbitmq-c library to be unloaded
-     * safely. We do leak the amqp_openssl_lockarray. Which is only
-     * an issue if you repeatedly unload and load the library
-     */
-    CRYPTO_set_locking_callback(NULL);
-    CRYPTO_set_id_callback(NULL);
-  }
-
-  pthread_mutex_unlock(&openssl_init_mutex);
-#endif /* ENABLE_THREAD_SAFETY */
-  return 0;
+  status = AMQP_STATUS_OK;
+out:
+  CHECK_SUCCESS(pthread_mutex_unlock(&openssl_init_mutex));
+  return status;
+}
+
+static int initialize_ssl_and_increment_connections() {
+  int status;
+  CHECK_SUCCESS(pthread_mutex_lock(&openssl_init_mutex));
+
+  if (do_initialize_openssl && !openssl_initialized) {
+    status = setup_openssl();
+    if (status) {
+      goto exit;
+    }
+    openssl_initialized = 1;
+  }
+
+  if (!openssl_bio_initialized) {
+    status = amqp_openssl_bio_init();
+    if (status) {
+      goto exit;
+    }
+    openssl_bio_initialized = 1;
+  }
+
+  openssl_connections += 1;
+  status = AMQP_STATUS_OK;
+exit:
+  CHECK_SUCCESS(pthread_mutex_unlock(&openssl_init_mutex));
+  return status;
+}
+
+static int decrement_ssl_connections(void) {
+  CHECK_SUCCESS(pthread_mutex_lock(&openssl_init_mutex));
+
+  if (openssl_connections > 0) {
+    openssl_connections--;
+  }
+
+  CHECK_SUCCESS(pthread_mutex_unlock(&openssl_init_mutex));
+  return AMQP_STATUS_OK;
+}
+
+int amqp_uninitialize_ssl_library(void) {
+  int status;
+  CHECK_SUCCESS(pthread_mutex_lock(&openssl_init_mutex));
+
+  if (openssl_connections > 0) {
+    status = AMQP_STATUS_SOCKET_INUSE;
+    goto out;
+  }
+
+  amqp_openssl_bio_destroy();
+  openssl_bio_initialized = 0;
+
+#ifndef AMQP_OPENSSL_V110
+  ERR_remove_state(0);
+#endif
+
+#ifndef LIBRESSL_VERSION_NUMBER
+  FIPS_mode_set(0);
+#endif
+
+  CRYPTO_set_locking_callback(NULL);
+  CRYPTO_set_id_callback(NULL);
+  {
+    int i;
+    for (i = 0; i < CRYPTO_num_locks(); i++) {
+      pthread_mutex_destroy(&amqp_openssl_lockarray[i]);
+    }
+    free(amqp_openssl_lockarray);
+  }
+
+  ENGINE_cleanup();
+  CONF_modules_free();
+  EVP_cleanup();
+  CRYPTO_cleanup_all_ex_data();
+  ERR_free_strings();
+#if (OPENSSL_VERSION_NUMBER >= 0x10002003L) && !defined(LIBRESSL_VERSION_NUMBER)
+  SSL_COMP_free_compression_methods();
+#endif
+
+  openssl_initialized = 0;
+
+  status = AMQP_STATUS_OK;
+out:
+  CHECK_SUCCESS(pthread_mutex_unlock(&openssl_init_mutex));
+  return status;
 }
